@@ -23,6 +23,7 @@
 #include <unistd.h>
 #include <getopt.h>
 #include <string.h>
+#include <stdarg.h>
 
 #include <sys/types.h>
 #include <sys/socket.h>
@@ -30,14 +31,14 @@
 #include <arpa/inet.h>
 #include <netdb.h>
 
-#include <signal.h> 
+#include <signal.h>
 
 #if HAVE_READLINE
 #include <readline/readline.h>
 #include <readline/history.h>
 #endif
 
-#if HAVE_ARGZ
+#if HAVE_ARGZ_H
 #include <argz.h>
 #else
 #include "argz.h"
@@ -57,6 +58,7 @@
 /* defined as extern in tftp_file.c and mtftp_file.c, set by the signal
    handler */
 int tftp_cancel = 0;
+int tftp_prevent_sas = 0;
 
 /* local flags */
 int interactive = 1;            /* if false, we run in batch mode */
@@ -139,7 +141,7 @@ int main(int argc, char **argv)
      data.data_buffer_size = SEGSIZE + 4;
 
      /* Allocate memory for tftp option structure. */
-     if ((data.tftp_options = 
+     if ((data.tftp_options =
           malloc(sizeof(tftp_default_options))) == NULL)
      {
           fprintf(stderr, "tftp: memory allocation failed.\n");
@@ -147,10 +149,10 @@ int main(int argc, char **argv)
      }
      /* Copy default options. */
      memcpy(data.tftp_options, tftp_default_options,
-            sizeof(tftp_default_options));   
+            sizeof(tftp_default_options));
 
      /* Allocate memory for tftp option reply from server. */
-     if ((data.tftp_options_reply = 
+     if ((data.tftp_options_reply =
           malloc(sizeof(tftp_default_options))) == NULL)
      {
           fprintf(stderr, "tftp: memory allocation failed.\n");
@@ -187,13 +189,13 @@ int main(int argc, char **argv)
      return OK;
 }
 
-/* 
+/*
  * When we receive a signal, we set tftp_cancel in order to
  * abort ongoing transfer.
  */
 void signal_handler(int signal)
 {
-     /* 
+     /*
       * if receiving or sending files, we should abort
       * and send and error ACK
       */
@@ -300,9 +302,9 @@ int getc_func(FILE *fp)
 char **completion(const char *text, int start, int end)
 {
      char **matches;
-     
+
      matches = (char **)NULL;
-     
+
      /* If this word is at the start of the line, then it is a command
         to complete.  Otherwise it is the name of a file in the current
         directory. */
@@ -323,7 +325,7 @@ char *command_generator(const char *text, int state)
 {
      static int list_index, len;
      char *name;
-     
+
      /* If this is a new word to complete, initialize now.  This
         includes saving the length of TEXT for efficiency, and
         initializing the index variable to 0. */
@@ -337,24 +339,49 @@ char *command_generator(const char *text, int state)
      while ((name = (char *)cmdtab[list_index].name))
      {
           list_index++;
-          
+
           if (strncmp (name, text, len) == 0)
                return strdup(name);
      }
-     
+
      /* If no names matched, then return NULL. */
      return NULL;
 }
 # endif
 #endif
 
+/*
+ * set argc/argv from variadic string arguments
+*/
+void make_arg_vector(int *argc, char***argv, ...)
+{
+  char **p;
+  char *s;
+  va_list argp;
+
+  // how many args?
+  *argc = 0;
+  va_start(argp, argv);
+  while ( (s=va_arg(argp, char*)) )
+    ++*argc;
+
+  // allocate storage
+  *argv = malloc(*argc * sizeof (char*));
+
+  // store args
+  p = *argv;
+  va_start(argp, argv);
+  while ( (s=va_arg(argp, char*)) )
+    *p++ = s;
+}
+
 /*
  * Split a string into args.
  */
 void make_arg(char *string, int *argc, char ***argv)
 {
      static char *tmp = NULL;
-     int argz_len;
+     size_t argz_len;
 
      /* split the string to an argz vector */
      if (argz_create_sep(string, ' ', &tmp, &argz_len) != 0)
@@ -372,7 +399,7 @@ void make_arg(char *string, int *argc, char ***argv)
      }
      /* extract arguments */
      argz_extract(tmp, argz_len, *argv);
-     
+
      /* if the last argument is an empty string ... it happens
         when some extra space are added at the end of string :( */
      if (strlen((*argv)[*argc - 1]) == 0)
@@ -408,8 +435,8 @@ int process_cmd(int argc, char **argv)
  */
 int set_peer(int argc, char **argv)
 {
-     struct hostent *host;      /* for host name lookup */
-     struct servent *sp;        /* server entry for tftp service */
+     struct addrinfo hints, *addrinfo;
+     int err;
 
      /* sanity check */
      if ((argc < 2) || (argc > 3))
@@ -418,47 +445,40 @@ int set_peer(int argc, char **argv)
           return ERR;
      }
 
-     /* get the server entry */
-     sp = getservbyname("tftp", "udp");
-     if (sp == 0) {
-          fprintf(stderr, "tftp: udp/tftp, unknown service.\n");
-          return ERR;
-     }
-
-     /* look up the host */
-     host = gethostbyname(argv[1]);
+     /* look up the service and host */
+     memset(&hints, 0, sizeof(hints));
+     hints.ai_socktype = SOCK_DGRAM;
+     hints.ai_flags = AI_CANONNAME;
+     err = getaddrinfo(argv[1], argc == 3 ? argv[2] : "tftp",
+                       &hints, &addrinfo);
      /* if valid, update s_inn structure */
-     if (host)
+     if (err == 0)
+          err = sockaddr_set_addrinfo(&data.sa_peer, addrinfo);
+     if (err == 0)
      {
-          data.sa_peer.sin_family = host->h_addrtype;
-          if (host->h_length > sizeof(data.sa_peer.sin_addr))
-               host->h_length = sizeof(data.sa_peer.sin_addr);
-          memcpy(&data.sa_peer.sin_addr, host->h_addr, host->h_length);
-          Strncpy(data.hostname, host->h_name,
+          Strncpy(data.hostname, addrinfo->ai_canonname,
                   sizeof(data.hostname));
           data.hostname[sizeof(data.hostname)-1] = 0;
-          data.sa_peer.sin_port = sp->s_port;
-     } 
+          freeaddrinfo(addrinfo);
+     }
      else
      {
-          fprintf(stderr, "tftp: unknown host %s.\n", argv[1]);
+          if (err == EAI_SERVICE)
+          {
+               if (argc == 3)
+                    fprintf(stderr, "%s: bad port number.\n", argv[2]);
+               else
+                    fprintf(stderr, "tftp: udp/tftp, unknown service.\n");
+          }
+          else
+          {
+               fprintf(stderr, "tftp: unknown host %s.\n", argv[1]);
+          }
           data.connected = 0;
           return ERR;
      }
-     /* get the server port */
-     if (argc == 3)
-     {
-          sp->s_port = htons(atoi(argv[2]));
-          if (sp->s_port < 0)
-          {
-               fprintf(stderr, "%s: bad port number.\n", argv[2]);
-               data.connected = 0;
-               return ERR;
-          }
-          data.sa_peer.sin_port = sp->s_port;
-     }
      /* copy port number to data structure */
-     data.port = ntohs(sp->s_port);
+     data.port = sockaddr_get_port(&data.sa_peer);
 
      data.connected = 1;
      return OK;
@@ -508,23 +528,32 @@ int set_option(int argc, char **argv)
           fprintf(stderr, "Usage: option <option name> [option value]\n");
           fprintf(stderr, "       option disable <option name>\n");
           if (data.tftp_options[OPT_TSIZE].specified)
-               fprintf(stderr, "  tsize:     enabled\n");
+               fprintf(stderr, "  tsize:      enabled\n");
           else
-               fprintf(stderr, "  tsize:     disabled\n");
+               fprintf(stderr, "  tsize:      disabled\n");
           if (data.tftp_options[OPT_BLKSIZE].specified)
-               fprintf(stderr, "  blksize:   %s\n",
+               fprintf(stderr, "  blksize:    %s\n",
                        data.tftp_options[OPT_BLKSIZE].value);
           else
-               fprintf(stderr, "  blksize:   disabled\n");
+               fprintf(stderr, "  blksize:    disabled\n");
+          if (data.tftp_options[OPT_WINDOWSIZE].specified)
+               fprintf(stderr, "  windowsize:   %s\n",
+                       data.tftp_options[OPT_WINDOWSIZE].value);
+          else
+               fprintf(stderr, "  windowsize: disabled\n");
           if (data.tftp_options[OPT_TIMEOUT].specified)
-               fprintf(stderr, "  timeout:   %s\n",
+               fprintf(stderr, "  timeout:    %s\n",
                        data.tftp_options[OPT_TIMEOUT].value);
           else
-               fprintf(stderr, "  timeout:   disabled\n");
+               fprintf(stderr, "  timeout:    disabled\n");
           if (data.tftp_options[OPT_MULTICAST].specified)
-               fprintf(stderr, "  multicast: enabled\n");
+               fprintf(stderr, "  multicast:  enabled\n");
           else
-               fprintf(stderr, "  multicast: disabled\n");
+               fprintf(stderr, "  multicast:  disabled\n");
+          if (data.tftp_options[OPT_PASSWORD].specified)
+               fprintf(stderr, "  password:   enabled\n");
+          else
+               fprintf(stderr, "  password:   disabled\n");
           return ERR;
      }
      /* if disabling an option */
@@ -600,9 +629,9 @@ int put_file(int argc, char **argv)
           Strncpy(data.local_file, argv[1], VAL_SIZE);
           Strncpy(data.tftp_options[OPT_FILENAME].value, argv[2], VAL_SIZE);
      }
-     
+
      /* open a UDP socket */
-     data.sockfd = socket(AF_INET, SOCK_DGRAM, 0);
+     data.sockfd = socket(data.sa_peer.ss_family, SOCK_DGRAM, 0);
      if (data.sockfd < 0) {
           perror("tftp: ");
           exit(ERR);
@@ -621,7 +650,7 @@ int put_file(int argc, char **argv)
      fsync(data.sockfd);
      close(data.sockfd);
 
-     return OK;
+     return tftp_result;
 }
 
 /*
@@ -685,11 +714,14 @@ int get_file(int argc, char **argv)
                {
                     fclose(fp);
 #if HAVE_READLINE
-                    string = readline("Overwite local file [y/n]? ");
+                    string = readline("Overwrite local file [y/n]? ");
 #else
-                    fprintf(stderr, "Overwite local file [y/n]? ");
-                    fgets(string, MAXLEN, stdin);
-                    string[strlen(string) - 1] = 0;
+                    fprintf(stderr, "Overwrite local file [y/n]? ");
+                    if (fgets(string, MAXLEN, stdin) == NULL) {
+                         string[0] = 0;
+                    } else {
+                         string[strlen(string) - 1] = 0;
+                    }
 #endif
                     if (!(strcasecmp(string, "y") == 0))
                     {
@@ -706,7 +738,7 @@ int get_file(int argc, char **argv)
      }
 
      /* open a UDP socket */
-     data.sockfd = socket(AF_INET, SOCK_DGRAM, 0);
+     data.sockfd = socket(data.sa_peer.ss_family, SOCK_DGRAM, 0);
      if (data.sockfd < 0) {
           perror("tftp: ");
           exit(ERR);
@@ -731,11 +763,11 @@ int get_file(int argc, char **argv)
      fsync(data.sockfd);
      close(data.sockfd);
 
-     return OK;
+     return tftp_result;
 }
 
 #ifdef HAVE_MTFTP
-/* 
+/*
  * Set ot get mtftp variable value
  */
 int mtftp_opt(int argc, char **argv)
@@ -836,34 +868,38 @@ int status(int argc, char **argv)
      if (!data.connected)
           fprintf(stderr, "Not connected\n");
      else
-          fprintf(stderr, "Connected:  %s port %d\n", data.hostname,
+          fprintf(stderr, "Connected:   %s port %d\n", data.hostname,
                   data.port);
      fprintf(stderr, "Mode:       %s\n", data.tftp_options[OPT_MODE].value);
      if (data.verbose)
-          fprintf(stderr, "Verbose:    on\n");
+          fprintf(stderr, "Verbose:     on\n");
      else
-          fprintf(stderr, "Verbose:    off\n");
+          fprintf(stderr, "Verbose:     off\n");
      if (data.trace)
-          fprintf(stderr, "Trace:      on\n");
+          fprintf(stderr, "Trace:       on\n");
      else
-          fprintf(stderr, "Trace:      off\n");
+          fprintf(stderr, "Trace:       off\n");
      fprintf(stderr, "Options\n");
      if (data.tftp_options[OPT_TSIZE].specified)
-          fprintf(stderr, " tsize:     enabled\n");
+          fprintf(stderr, " tsize:      enabled\n");
      else
-          fprintf(stderr, " tsize:     disabled\n");
+          fprintf(stderr, " tsize:      disabled\n");
      if (data.tftp_options[OPT_BLKSIZE].specified)
-          fprintf(stderr, " blksize:   enabled\n");
+          fprintf(stderr, " blksize:    enabled\n");
      else
-          fprintf(stderr, " blksize:   disabled\n");
+          fprintf(stderr, " blksize:    disabled\n");
+     if (data.tftp_options[OPT_WINDOWSIZE].specified)
+          fprintf(stderr, " windowsize: enabled\n");
+     else
+          fprintf(stderr, " windowsize: disabled\n");
      if (data.tftp_options[OPT_TIMEOUT].specified)
-          fprintf(stderr, " timeout:   enabled\n");
+          fprintf(stderr, " timeout:    enabled\n");
      else
-          fprintf(stderr, " timeout:   disabled\n");
+          fprintf(stderr, " timeout:    disabled\n");
      if (data.tftp_options[OPT_MULTICAST].specified)
-          fprintf(stderr, " multicast: enabled\n");
+          fprintf(stderr, " multicast:  enabled\n");
      else
-          fprintf(stderr, " multicast: disabled\n");
+          fprintf(stderr, " multicast:  disabled\n");
 #ifdef HAVE_MTFTP
      fprintf(stderr, "mtftp variables\n");
      fprintf(stderr, " client-port:   %d\n", data.mtftp_client_port);
@@ -885,7 +921,7 @@ int status(int argc, char **argv)
           if (tftp_result == OK)
           {
                print_eng((double)data.file_size, string, sizeof(string), "%3.3f%cB");
-               fprintf(stderr, "  Bytes transfered:  %s\n", string);
+               fprintf(stderr, "  Bytes transferred:  %s\n", string);
                fprintf(stderr, "  Time of transfer: %8.3fs\n",
                        (double)(tmp.tv_sec + tmp.tv_usec * 1e-6));
                fprintf(stderr, "  Throughput:        ");
@@ -895,7 +931,7 @@ int status(int argc, char **argv)
                fprintf(stderr, "%s", string);
           }
           else
-               fprintf(stderr, "  Transfer aborted\n"); 
+               fprintf(stderr, "  Transfer aborted\n");
      }
      return OK;
 }
@@ -958,8 +994,8 @@ int tftp_cmd_line_options(int argc, char **argv)
      int ac;                    /* to format arguments for process_cmd */
      char **av = NULL;          /* same */
      char string[MAXLEN];
-     char local_file[MAXLEN] = " ";
-     char remote_file[MAXLEN] = " ";
+     char local_file[MAXLEN] = "";
+     char remote_file[MAXLEN] = "";
      int action = 0;
 
      int option_index = 0;
@@ -971,6 +1007,7 @@ int tftp_cmd_line_options(int argc, char **argv)
           { "put", 0, NULL, 'p'},
           { "local-file", 1, NULL, 'l'},
           { "remote-file", 1, NULL, 'r'},
+          { "password", 1, NULL, 'P'},
           { "tftp-timeout", 1, NULL, 'T'},
           { "mode", 1, NULL, 'M'},
           { "option", 1, NULL, 'O'},
@@ -982,6 +1019,7 @@ int tftp_cmd_line_options(int argc, char **argv)
 #endif
           { "mtftp", 1, NULL, '1'},
           { "no-source-port-checking", 0, NULL, '0'},
+          { "prevent-sas", 0, NULL, 'X'},
           { "verbose", 0, NULL, 'v'},
           { "trace", 0, NULL, 'd'},
 #if DEBUG
@@ -993,7 +1031,7 @@ int tftp_cmd_line_options(int argc, char **argv)
      };
 
      /* Support old argument until 0.8 */
-     while ((c = getopt_long(argc, argv, /*"gpl:r:Vh"*/ "gpl:r:Vht:b:sm",
+     while ((c = getopt_long(argc, argv, /*"gpl:r:Vh"*/ "gpl:r:Vht:b:smP:",
                              options, &option_index)) != EOF)
      {
           switch (c)
@@ -1028,6 +1066,11 @@ int tftp_cmd_line_options(int argc, char **argv)
                else
                     action = PUT;
                break;
+          case 'P':
+               snprintf(string, sizeof(string), "option password %s", optarg);
+               make_arg(string, &ac, &av);
+               process_cmd(ac, av);
+               break;
           case 'l':
                interactive = 0;
                Strncpy(local_file, optarg, MAXLEN);
@@ -1058,26 +1101,26 @@ int tftp_cmd_line_options(int argc, char **argv)
                break;
 #if 1
           case 't':
-               fprintf(stderr, "--timeout deprecate, use --option instead\n");
+               fprintf(stderr, "--timeout deprecated, use --option instead\n");
                snprintf(string, sizeof(string), "option timeout %s", optarg);
                make_arg(string, &ac, &av);
                process_cmd(ac, av);
                break;
           case 'b':
-               fprintf(stderr, "--blksize deprecate, use --option instead\n");
+               fprintf(stderr, "--blksize deprecated, use --option instead\n");
                snprintf(string, sizeof(string), "option blksize %s", optarg);
                make_arg(string, &ac, &av);
                process_cmd(ac, av);
                break;
           case 's':
-               fprintf(stderr, "--tsize deprecate, use --option instead\n");
+               fprintf(stderr, "--tsize deprecated, use --option instead\n");
                snprintf(string, sizeof(string), "option tsize");
                make_arg(string, &ac, &av);
                process_cmd(ac, av);
                break;
           case 'm':
                fprintf(stderr,
-                       "--multicast deprecate, use --option instead\n");
+                       "--multicast deprecated, use --option instead\n");
                snprintf(string, sizeof(string), "option multicast");
                make_arg(string, &ac, &av);
                process_cmd(ac, av);
@@ -1086,15 +1129,18 @@ int tftp_cmd_line_options(int argc, char **argv)
           case '0':
                data.checkport = 0;
                break;
+          case 'X':
+               tftp_prevent_sas = 1;
+               break;
           case 'v':
                snprintf(string, sizeof(string), "verbose on");
                make_arg(string, &ac, &av);
-               process_cmd(ac, av);            
+               process_cmd(ac, av);
                break;
           case 'd':
                snprintf(string, sizeof(string), "trace on");
                make_arg(string, &ac, &av);
-               process_cmd(ac, av);            
+               process_cmd(ac, av);
                break;
 #if DEBUG
           case 'D':
@@ -1133,24 +1179,38 @@ int tftp_cmd_line_options(int argc, char **argv)
           make_arg(string, &ac, &av);
           process_cmd(ac, av);
      }
-     
+
      if (!interactive)
      {
           if (action == PUT)
-               snprintf(string, sizeof(string), "put %s %s", local_file,
-                        remote_file);
+          {
+               if(strlen(remote_file) == 0)
+               {
+                   strncpy(remote_file, local_file, MAXLEN);
+               }
+               make_arg_vector(&ac,&av,"put",local_file,remote_file,NULL);
+          }
           else if (action == GET)
-               snprintf(string, sizeof(string), "get %s %s", remote_file,
-                        local_file);
+          {
+               if(strlen(local_file) == 0)
+               {
+                   strncpy(local_file, remote_file, MAXLEN);
+               }
+               make_arg_vector(&ac,&av,"get",remote_file,local_file,NULL);
+          }
           else if (action == MGET)
-               snprintf(string, sizeof(string), "mget %s %s", remote_file,
-                        local_file);
+          {
+               if(strlen(local_file) == 0)
+               {
+                   strncpy(local_file, remote_file, MAXLEN);
+               }
+               make_arg_vector(&ac,&av,"mget",remote_file,local_file,NULL);
+          }
           else
           {
                fprintf(stderr, "No action specified in batch mode!\n");
                exit(ERR);
           }
-          make_arg(string, &ac, &av);
           if (process_cmd(ac, av) == ERR)
                exit(ERR);
      }
@@ -1169,6 +1229,7 @@ void tftp_usage(void)
              "  -p, --put                : put file\n"
              "  -l, --local-file <file>  : local file name\n"
              "  -r, --remote-file <file> : remote file name\n"
+             "  -P, --password <password>: specify password (Linksys extension)\n"
              "  --tftp-timeout <value>   : delay before retransmission, client side\n"
 #if 0
              "  t, --timeout <value>      : delay before retransmission, "
@@ -1182,6 +1243,7 @@ void tftp_usage(void)
              "  --mtftp <\"name value\">   : set mtftp variable to value\n"
 #endif
              "  --no-source-port-checking: violate RFC, see man page\n"
+             "  --prevent-sas            : prevent Sorcerer's Apprentice Syndrome\n"
              "  --verbose                : set verbose mode on\n"
              "  --trace                  : set trace mode on\n"
 #if DEBUG
@@ -1191,6 +1253,6 @@ void tftp_usage(void)
              "  -h, --help               : print this help\n"
              "\n"
              " [host] is the tftp server name\n"
-             " [port] is the port to use\n" 
-             "\n");     
+             " [port] is the port to use\n"
+             "\n");
 }
