@@ -55,6 +55,7 @@
 extern char directory[MAXLEN];
 /* read only except for the main thread */
 extern int tftpd_cancel;
+extern int tftpd_prevent_sas;
 
 #ifdef HAVE_PCRE
 extern tftpd_pcre_self_t *pcre_top;
@@ -75,7 +76,7 @@ int tftpd_rules_check(char *filename)
      {
           /* Append the prefix directory to the requested file. */
           snprintf(string, sizeof(string), "%s%s",
-                   directory, filename); 
+                   directory, filename);
           Strncpy(filename, string, MAXLEN);
      }
 
@@ -84,7 +85,7 @@ int tftpd_rules_check(char *filename)
      {
           /* Illegal access. */
           logger(LOG_INFO, "File name with /../ are forbidden");
-          return ERR;  
+          return ERR;
      }
      return OK;
 }
@@ -98,7 +99,7 @@ int tftpd_rules_check(char *filename)
  *
  *     1) send a ACK or OACK
  *     2) wait replay
- *          - if DATA packet, read it, send an acknoledge, goto 2
+ *          - if DATA packet, read it, send an acknowledge, goto 2
  *          - if ERROR abort
  *          - if TIMEOUT goto previous state
  */
@@ -106,30 +107,34 @@ int tftpd_receive_file(struct thread_data *data)
 {
      int state = S_BEGIN;
      int timeout_state = state;
+     int windowblock = 0;         /* number of block in a window, c.f. RFC7440 */
+     int windowsize = 1;          /* c.f. RFC7440 */
      int result;
      long block_number = 0;
+     long last_received_block = 0;
+     int extra_ack = 0;           /* number of additional ACKs */
      int data_size;
      int sockfd = data->sockfd;
      struct sockaddr_storage *sa = &data->client_info->client;
      struct sockaddr_storage from;
      char addr_str[SOCKADDR_PRINT_ADDR_LEN];
      struct tftphdr *tftphdr = (struct tftphdr *)data->data_buffer;
-     FILE *fp;
+     FILE *fp = NULL;
      char filename[MAXLEN];
      char string[MAXLEN];
      int timeout = data->timeout;
      int number_of_timeout = 0;
      int all_blocks_received = 0; /* temporary kludge */
-     int convert = 0;           /* if true, do netascii convertion */
+     int convert = 0;           /* if true, do netascii conversion */
 
-     long prev_block_number = 0; /* needed to support netascii convertion */
+     long prev_block_number = 0; /* needed to support netascii conversion */
      int temp = 0;
 
      /* look for mode option */
      if (strcasecmp(data->tftp_options[OPT_MODE].value, "netascii") == 0)
      {
           convert = 1;
-          logger(LOG_DEBUG, "will do netascii convertion");
+          logger(LOG_DEBUG, "will do netascii conversion");
      }
 
      /* file name verification */
@@ -144,18 +149,6 @@ int tftpd_receive_file(struct thread_data *data)
           return ERR;
      }
 
-     /* Open the file for writing. */
-     if ((fp = fopen(filename, "w")) == NULL)
-     {
-          /* Can't create the file. */
-          logger(LOG_INFO, "Can't open %s for writing", filename);
-          tftp_send_error(sockfd, sa, EACCESS, data->data_buffer, data->data_buffer_size);
-          if (data->trace)
-               logger(LOG_DEBUG, "sent ERROR <code: %d, msg: %s>", EACCESS,
-                      tftp_errmsg[EACCESS]);
-          return ERR;
-     }
-
      /* tsize option */
      if (((result = opt_get_tsize(data->tftp_options)) > -1) && !convert)
      {
@@ -163,6 +156,14 @@ int tftpd_receive_file(struct thread_data *data)
           logger(LOG_DEBUG, "tsize option -> %d", result);
      }
 
+     /* windowsize */
+     if ((result = opt_get_windowsize(data->tftp_options)) > -1)
+     {
+          windowsize = result;
+          opt_set_windowsize(result, data->tftp_options);
+          logger(LOG_DEBUG, "windowsize option -> %d", result);
+     }
+
      /* timeout option */
      if ((result = opt_get_timeout(data->tftp_options)) > -1)
      {
@@ -172,7 +173,6 @@ int tftpd_receive_file(struct thread_data *data)
                if (data->trace)
                     logger(LOG_DEBUG, "sent ERROR <code: %d, msg: %s>", EOPTNEG,
                            tftp_errmsg[EOPTNEG]);
-               fclose(fp);
                return ERR;
           }
           timeout = result;
@@ -180,16 +180,24 @@ int tftpd_receive_file(struct thread_data *data)
           logger(LOG_DEBUG, "timeout option -> %d", timeout);
      }
 
-     /* blksize options */
+     /* blksize */
      if ((result = opt_get_blksize(data->tftp_options)) > -1)
      {
-          if ((result < 8) || (result > 65464))
+          /*
+           *  If we receive more options, we have to make sure our buffer for
+           *  the OACK is not too small.  Use the string representation of
+           *  the options here for simplicity, which puts us on the safe side.
+           *  FIXME: Use independent buffers for OACK and data.
+           */
+          opt_options_to_string(data->tftp_options, string, MAXLEN);
+          if ((result < strlen(string)-2) || (result > 65464))
           {
+               logger(LOG_NOTICE, "options <%s> require roughly a blksize of %d for the OACK.",
+                      string, strlen(string)-2);
                tftp_send_error(sockfd, sa, EOPTNEG, data->data_buffer, data->data_buffer_size);
                if (data->trace)
                     logger(LOG_DEBUG, "sent ERROR <code: %d, msg: %s>", EOPTNEG,
                            tftp_errmsg[EOPTNEG]);
-               fclose(fp);
                return ERR;
           }
 
@@ -199,7 +207,6 @@ int tftpd_receive_file(struct thread_data *data)
           if (data->data_buffer == NULL)
           {
                logger(LOG_ERR, "memory allocation failure");
-               fclose(fp);
                return ERR;
           }
           tftphdr = (struct tftphdr *)data->data_buffer;
@@ -210,7 +217,6 @@ int tftpd_receive_file(struct thread_data *data)
                if (data->trace)
                     logger(LOG_DEBUG, "sent ERROR <code: %d, msg: %s>", ENOSPACE,
                            tftp_errmsg[ENOSPACE]);
-               fclose(fp);
                return ERR;
           }
           opt_set_blksize(result, data->tftp_options);
@@ -240,29 +246,32 @@ int tftpd_receive_file(struct thread_data *data)
                     state = S_SEND_ACK;
                break;
           case S_SEND_ACK:
+               windowblock = 0;
                timeout_state = state;
-               tftp_send_ack(sockfd, sa, block_number);
+               tftp_send_ack(sockfd, sa, last_received_block);
                if (data->trace)
-                    logger(LOG_DEBUG, "sent ACK <block: %ld>", block_number);
+                    logger(LOG_DEBUG, "sent ACK <block: %ld>", last_received_block);
                if (all_blocks_received)
                     state = S_END;
                else
                     state = S_WAIT_PACKET;
                break;
           case S_SEND_OACK:
+               windowblock = 0;
                timeout_state = state;
                tftp_send_oack(sockfd, sa, data->tftp_options,
                               data->data_buffer, data->data_buffer_size);
-               opt_options_to_string(data->tftp_options, string, MAXLEN);
-               if (data->trace)
+               if (data->trace) {
+                    opt_options_to_string(data->tftp_options, string, MAXLEN);
                     logger(LOG_DEBUG, "sent OACK <%s>", string);
+               }
                state = S_WAIT_PACKET;
                break;
           case S_WAIT_PACKET:
                data_size = data->data_buffer_size;
                result = tftp_get_packet(sockfd, -1, NULL, sa, &from, NULL,
-                                        timeout, &data_size, data->data_buffer);
-               
+                                        timeout, 0, &data_size, data->data_buffer);
+
                switch (result)
                {
                case GET_TIMEOUT:
@@ -276,7 +285,7 @@ int tftpd_receive_file(struct thread_data *data)
                     }
                     else
                     {
-                         logger(LOG_WARNING, "timeout: retrying...");
+                         logger(LOG_WARNING, "timeout: retrying ...");
                          state = timeout_state;
                     }
                     break;
@@ -345,16 +354,54 @@ int tftpd_receive_file(struct thread_data *data)
                }
                break;
           case S_DATA_RECEIVED:
-               /* We need to seek to the right place in the file */
-	       block_number = tftp_rollover_blocknumber(
-		      ntohs(tftphdr->th_block), prev_block_number, 0);
-               if (data->trace)
-                    logger(LOG_DEBUG, "received DATA <block: %ld, size: %d>",
-                           block_number, data_size - 4);
+               if (fp == NULL) {
+                       /* Open the file for writing. */
+                       if ((fp = fopen(filename, "w")) == NULL)
+                       {
+                               /* Can't create the file. */
+                               logger(LOG_INFO, "Can't open %s for writing", filename);
+                               tftp_send_error(sockfd, sa, EACCESS, data->data_buffer, data->data_buffer_size);
+                               if (data->trace)
+                                       logger(LOG_DEBUG, "sent ERROR <code: %d, msg: %s>", EACCESS,
+                                                       tftp_errmsg[EACCESS]);
+                               return ERR;
+                       }
+               }
 
+               /* We need to seek to the right place in the file */
+               block_number = tftp_rollover_blocknumber(ntohs(tftphdr->th_block), prev_block_number, 0);
+               /* The first data package acknowledges the OACK */
+               if ((last_received_block < windowsize) && (timeout_state = S_SEND_OACK))
+                    timeout_state = S_SEND_ACK;
+
+               /* The sequence of blocks may contain a gap, i.e. (block_number > last_received_block + 1).
+                  Send an extra ACK for the last block before the gap once.
+
+                  The ACK of a received window gets lost.  The sender times out and starts
+                  sending all blocks of that window again, i.e. (block_number < last_received_block + 1).
+                  Again, send the extra ACK. */
+               if (block_number != last_received_block + 1) {
+                    /* We got a wrong block from the client, send ACK for last block again */
+                    if (extra_ack < 1) {
+                         logger(LOG_DEBUG, "got wrong block <block: %ld>, sending extra ACK for <block: %ld>",
+                                block_number, last_received_block);
+                         extra_ack++;
+                         state = S_SEND_ACK;
+                    } else {
+                         logger(LOG_DEBUG, "got wrong block <block: %ld>, ignoring", block_number);
+                         state = S_WAIT_PACKET;
+                    }
+                    break;
+               }
+               extra_ack = 0;
+
+               windowblock++;
+               if (data->trace)
+                    logger(LOG_DEBUG, "received %d. DATA <block: %ld, size %d>, update last received block: %ld → %ld>",
+                           windowblock, block_number, data_size - 4, last_received_block, block_number);
+               last_received_block = block_number;
                if (tftp_file_write(fp, tftphdr->th_data, data->data_buffer_size - 4, block_number,
-                                   data_size - 4, convert, &prev_block_number, &temp)
-                   != data_size - 4)
+                                   data_size - 4, convert, &prev_block_number, &temp) != data_size - 4)
                {
                     logger(LOG_ERR, "%s: %d: error writing to file %s",
                            __FILE__, __LINE__, filename);
@@ -370,16 +417,21 @@ int tftpd_receive_file(struct thread_data *data)
                     all_blocks_received = 1;
                else
                     all_blocks_received = 0;
-               state = S_SEND_ACK;
+               if ((windowblock >= windowsize) || (all_blocks_received))
+                    state = S_SEND_ACK;
+               else
+                    state = S_WAIT_PACKET;
                break;
           case S_END:
-               fclose(fp);
+               if (fp != NULL) fclose(fp);
+               logger(LOG_DEBUG, "End of transfer");
                return OK;
           case S_ABORT:
-               fclose(fp);
+               if (fp != NULL) fclose(fp);
+               logger(LOG_DEBUG, "Aborting transfer");
                return ERR;
           default:
-               fclose(fp);
+               if (fp != NULL) fclose(fp);
                logger(LOG_ERR, "%s: %d: tftpd_file.c: huh?",
                       __FILE__, __LINE__);
                return ERR;
@@ -407,6 +459,7 @@ int tftpd_send_file(struct thread_data *data)
      int timeout_state = state;
      int result;
      long block_number = 0;
+     long last_ackd_block = -1;
      long last_block = -1;
      int data_size;
      struct sockaddr_storage *sa = &data->client_info->client;
@@ -418,6 +471,10 @@ int tftpd_send_file(struct thread_data *data)
      char filename[MAXLEN];
      char string[MAXLEN];
      int timeout = data->timeout;
+     int windowblock = 0;
+     int windowsize = 1;
+     int in_window_ack = -1;
+     int delay = 0;             /* delay between datagrams in a window */
      int number_of_timeout = 0;
      int mcast_switch = data->mcast_switch_client;
      struct stat file_stat;
@@ -430,15 +487,19 @@ int tftpd_send_file(struct thread_data *data)
      struct client_info *client_old = NULL;
      struct tftp_opt options[OPT_NUMBER];
 
-     long prev_block_number = 0; /* needed to support netascii convertion */
+     long prev_block_number = 0; /* needed to support netascii conversion */
      long prev_file_pos = 0;
      int temp = 0;
 
+     int prev_sent_count = 0;
+     int prev_ack_count = 0;
+     int curr_sent_count = 0;
+
      /* look for mode option */
      if (strcasecmp(data->tftp_options[OPT_MODE].value, "netascii") == 0)
      {
           convert = 1;
-          logger(LOG_DEBUG, "will do netascii convertion");
+          logger(LOG_DEBUG, "will do netascii conversion");
      }
 
      /* file name verification */
@@ -469,7 +530,7 @@ int tftpd_send_file(struct thread_data *data)
                }
                else
                {
-                    logger(LOG_INFO, "PCRE mapped %s -> %s", 
+                    logger(LOG_INFO, "PCRE mapped %s -> %s",
                            data->tftp_options[OPT_FILENAME].value, string);
                     Strncpy(filename, string, MAXLEN);
                     /* recheck those rules */
@@ -527,11 +588,37 @@ int tftpd_send_file(struct thread_data *data)
           logger(LOG_INFO, "timeout option -> %d", timeout);
      }
 
-     /* blksize options */
+     /* windowsize option */
+     if ((result = opt_get_windowsize(data->tftp_options)) > -1)
+     {
+          if ((result < 1) || (result > 65535))
+          {
+               tftp_send_error(sockfd, sa, EOPTNEG, data->data_buffer, data->data_buffer_size);
+               if (data->trace)
+                    logger(LOG_DEBUG, "sent ERROR <code: %d, msg: %s>", EOPTNEG,
+                           tftp_errmsg[EOPTNEG]);
+               fclose(fp);
+               return ERR;
+          }
+          windowsize = result;
+          opt_set_windowsize(windowsize, data->tftp_options);
+          logger(LOG_INFO, "windowsize option -> %d", windowsize);
+     }
+
+     /* blksize */
      if ((result = opt_get_blksize(data->tftp_options)) > -1)
      {
-          if ((result < 8) || (result > 65464))
+          /*
+           *  If we receive more options, we have to make sure our buffer for
+           *  the OACK is not too small.  Use the string representation of
+           *  the options here for simplicity, which puts us on the safe side.
+           *  FIXME: Use independent buffers for OACK and data.
+           */
+          opt_options_to_string(data->tftp_options, string, MAXLEN);
+          if ((result < strlen(string)-2) || (result > 65464))
           {
+               logger(LOG_NOTICE, "options <%s> require roughly a blksize of %d for the OACK.",
+                      string, strlen(string)-2);
                tftp_send_error(sockfd, sa, EOPTNEG, data->data_buffer, data->data_buffer_size);
                if (data->trace)
                     logger(LOG_DEBUG, "sent ERROR <code: %d, msg: %s>", EOPTNEG,
@@ -624,14 +711,16 @@ int tftpd_send_file(struct thread_data *data)
                   not be a problem: the client thread will receive a second OACK and fall
                   back to non master mode. Then the server will timeout and either resend
                   OACK or continu with the next client */
-               opt_options_to_string(data->tftp_options, string, MAXLEN);
                if (data->trace)
+               {
+                    opt_options_to_string(data->tftp_options, string, MAXLEN);
                     logger(LOG_DEBUG, "sent OACK <%s>", string);
+               }
                tftp_send_oack(thread->sockfd, sa, data->tftp_options,
                               data->data_buffer, data->data_buffer_size);
 
                /* We are done */
-               logger(LOG_INFO, "Client transfered to %p", thread);
+               logger(LOG_INFO, "Client transferred to %p", thread);
                fclose(fp);
                return OK;
           }
@@ -681,13 +770,13 @@ int tftpd_send_file(struct thread_data *data)
                else
                     setsockopt(data->sockfd, IPPROTO_IPV6, IPV6_MULTICAST_HOPS,
                                &data->mcast_ttl, sizeof(data->mcast_ttl));
-               
+
                /* set options data for OACK */
                opt_set_multicast(data->tftp_options, data->mc_addr,
                                  data->mc_port, 1);
                logger(LOG_INFO, "multicast option -> %s,%d,%d", data->mc_addr,
                       data->mc_port, 1);
-            
+
                /* the socket must be unconnected for multicast */
                sa->ss_family = AF_UNSPEC;
                connect(sockfd, (struct sockaddr *)sa, sizeof(*sa));
@@ -736,16 +825,17 @@ int tftpd_send_file(struct thread_data *data)
                break;
           case S_SEND_OACK:
                timeout_state = state;
-               opt_options_to_string(data->tftp_options, string, MAXLEN);
-               if (data->trace)
+               if (data->trace) {
+                    opt_options_to_string(data->tftp_options, string, MAXLEN);
                     logger(LOG_DEBUG, "sent OACK <%s>", string);
+               }
                tftp_send_oack(sockfd, sa, data->tftp_options,
                               data->data_buffer, data->data_buffer_size);
                state = S_WAIT_PACKET;
                break;
           case S_SEND_DATA:
+               windowblock++;
                timeout_state = state;
-
                data_size = tftp_file_read(fp, tftphdr->th_data, data->data_buffer_size - 4, block_number,
                                           convert, &prev_block_number, &prev_file_pos, &temp);
                data_size += 4;  /* need to consider tftp header */
@@ -766,19 +856,37 @@ int tftpd_send_file(struct thread_data *data)
                                    data_size, data->data_buffer);
                }
                if (data->trace)
-                    logger(LOG_DEBUG, "sent DATA <block: %ld, size %d>",
-                           block_number + 1, data_size - 4);
+                    logger(LOG_DEBUG, "sent %d. DATA <block: %ld, size %d>",
+                           windowblock, block_number + 1, data_size - 4);
                state = S_WAIT_PACKET;
                break;
           case S_WAIT_PACKET:
                data_size = data->data_buffer_size;
-               result = tftp_get_packet(sockfd, -1, NULL, sa, &from, NULL,
-                                        timeout, &data_size, data->data_buffer);
+               if ((windowblock >= windowsize) || (last_ackd_block == -1) || (last_block != -1)) {
+                    /* we wait for the ACK */
+                    result = tftp_get_packet(sockfd, -1, NULL, sa, &from, NULL,
+                                             timeout, 0, &data_size, data->data_buffer);
+                    if ((delay > 0) && (in_window_ack == 0))
+                         delay--;
+                    in_window_ack = 0;
+               } else {
+                    /* we check if an unsolicitated ACK arrived */
+                    result = tftp_get_packet(sockfd, -1, NULL, sa, &from, NULL,
+                                             0, delay, &data_size, data->data_buffer);
+                    if (result == GET_TIMEOUT) {
+                         /* we send the next block */
+                         block_number = tftp_rollover_blocknumber(
+                              ntohs(tftphdr->th_block), prev_block_number, 0);
+                         state = S_SEND_DATA;
+                         break;
+                    }
+                    in_window_ack++;
+               }
                switch (result)
                {
                case GET_TIMEOUT:
                     number_of_timeout++;
-                    
+                    windowblock = 0;
                     if (number_of_timeout > NB_OF_RETRY)
                     {
                          logger(LOG_INFO, "client (%s) not responding",
@@ -801,10 +909,12 @@ int tftpd_send_file(struct thread_data *data)
                               {
                                    /* Send an OACK to the old client remove is
                                       master client status */
-                                   opt_options_to_string(options,
-                                                         string, MAXLEN);
+
                                    if (data->trace)
+                                   {
+                                        opt_options_to_string(options, string, MAXLEN);
                                         logger(LOG_DEBUG, "sent OACK <%s>", string);
+                                   }
                                    tftp_send_oack(sockfd, sa, options,
                                                   data->data_buffer, data->data_buffer_size);
 
@@ -818,6 +928,10 @@ int tftpd_send_file(struct thread_data *data)
                                           sockaddr_get_port(
                                                &client_info->client));
                                    sa = &client_info->client;
+
+                                   /* rewind the last_ackd_block counter */
+                                   last_ackd_block = -1;
+
                                    state = S_SEND_OACK;
                                    break;
                               }
@@ -831,7 +945,9 @@ int tftpd_send_file(struct thread_data *data)
                                    break;
                               }
                          }
-                         logger(LOG_WARNING, "timeout: retrying...");
+                         logger(LOG_WARNING, "timeout block %ld: retrying ...", block_number);
+                         /* rewind the block_number to the last block we received */
+                         block_number = last_ackd_block;
                          state = timeout_state;
                     }
                     break;
@@ -862,15 +978,15 @@ int tftpd_send_file(struct thread_data *data)
                               else
                                    /* If not, send and OACK with mc=0 to shut it up. */
                               {
-                                   opt_options_to_string(options,
-                                                         string, MAXLEN);
                                    if (data->trace)
+                                   {
+                                        opt_options_to_string(options, string, MAXLEN);
                                         logger(LOG_DEBUG, "sent OACK <%s>", string);
+                                   }
                                    tftp_send_oack(sockfd, &from, options,
                                                   data->data_buffer, data->data_buffer_size);
                               }
                               break;
-                              
                          }
                     }
                     else
@@ -887,27 +1003,112 @@ int tftpd_send_file(struct thread_data *data)
                                    break;
                               }
                               else
-                                   logger(LOG_WARNING,
-                                          "source port mismatch, check bypassed");
+                                   logger(LOG_WARNING, "source port mismatch, check bypassed");
                          }
                     }
+
                     /* The ACK is from the current client */
                     number_of_timeout = 0;
-		    if (multicast)
-			    block_number = ntohs(tftphdr->th_block);
-		    else
-		    {
-			    block_number = tftp_rollover_blocknumber(
-				ntohs(tftphdr->th_block), prev_block_number, 0);
-		    }
+                    if (multicast) {
+                         block_number = ntohs(tftphdr->th_block);
+                    } else {
+                         block_number = tftp_rollover_blocknumber(ntohs(tftphdr->th_block), prev_block_number, 0);
+                         // The last block has been sent, but not received. Reset last_block:
+                         if (block_number < last_block) last_block = -1;
+                    }
                     if (data->trace)
-                         logger(LOG_DEBUG, "received ACK <block: %ld>",
-                                block_number);
-                    if ((last_block != -1) && (block_number > last_block))
-                    {
+                         logger(LOG_DEBUG, "received ACK <block: %ld>", block_number);
+
+                    /* Now check the ACK number and possibly ignore the request */
+                    if (multicast) {
+                         /* multicast, block numbers could contain gaps */
+                         /* check whether the block request isn't already fulfilled */
+                         if (data->tftp_options[OPT_WINDOWSIZE].enabled || tftpd_prevent_sas) {
+                              if ((last_ackd_block >= block_number) || (in_window_ack > 1)) {
+                                   if (data->trace)
+                                        logger(LOG_DEBUG, "received outdated/duplicate ACK block: %ld", block_number);
+                                   break;
+                              } else
+                                   last_ackd_block = block_number;
+                         } else {
+                              /* don't prevent the SAS */
+                              /* use a heuristic suggested by Vladimir Nadvornik */
+                              /* here comes the ACK again */
+                              if (last_ackd_block == block_number) {
+                                   /* drop if number of ACKs == times of previous block sending */
+                                   if (++prev_ack_count == prev_sent_count) {
+                                        logger(LOG_DEBUG, "ACK count (%d) == previous block transmission count -> dropping ACK", prev_ack_count);
+                                        break;
+                                   }
+                                   /* else resend the block */
+                                   logger(LOG_DEBUG, "resending block %d", block_number + 1);
+                              }
+                              /* received ACK to sent block -> move on to next block */
+                              else if (last_ackd_block < block_number) {
+                                   last_ackd_block = block_number;
+                                   prev_sent_count = curr_sent_count;
+                                   curr_sent_count = 0;
+                                   prev_ack_count = 1;
+                              }
+                              /* block with low number -> ignore it completely */
+                              else {
+                                   logger(LOG_DEBUG, "ignoring ACK %d", block_number);
+                                   break;
+                              }
+                         }
+                    } else {
+                         /* unicast, blocks should be requested one after another */
+                         if (data->tftp_options[OPT_WINDOWSIZE].enabled || tftpd_prevent_sas) {
+                              /* check whether the block request isn't already fulfilled */
+                              if ((last_ackd_block > block_number) || (in_window_ack > 1)) {
+                                   if (data->trace)
+                                        logger(LOG_DEBUG, "ignore outdated/duplicate ACK: %ld", block_number);
+                                   break;
+                              }
+                              if ((last_ackd_block + windowsize != block_number) &&
+                                  (last_ackd_block != -1) && (last_block == -1)) {
+                                   logger(LOG_WARNING, "window block %ld not complete/ordered: %ld/%d valid, delay: %d usec",
+                                          block_number, block_number - last_ackd_block, windowsize, delay);
+                                   if (delay < timeout * 1000 / windowsize)
+                                        delay += 10;
+                              }
+                              if (data->trace)
+                                   logger(LOG_DEBUG, "update last ACK'd <block: %ld → %ld>", last_ackd_block, block_number);
+                              last_ackd_block = block_number;
+                         } else {
+                              /* don't prevent the SAS */
+                              /* use a heuristic suggested by Vladimir Nadvornik */
+                              /* here comes the ACK again */
+                              if (last_ackd_block == block_number) {
+                                   /* drop if number of ACKs == times of previous block sending */
+                                   if (++prev_ack_count == prev_sent_count) {
+                                        logger(LOG_DEBUG, "ACK count (%d) == previous block transmission count -> dropping ACK", prev_ack_count);
+                                        break;
+                                   }
+                                   /* else resend the block */
+                                   logger(LOG_DEBUG, "resending block %d", block_number + 1);
+                              }
+                              /* received ACK to sent block -> move on to next block */
+                              else if (last_ackd_block < block_number) {
+                                   last_ackd_block = block_number;
+                                   prev_sent_count = curr_sent_count;
+                                   curr_sent_count = 0;
+                                   prev_ack_count = 1;
+                              }
+                              /* nor previous nor current block number -> ignore it completely */
+                              else {
+                                   logger(LOG_DEBUG, "ignoring ACK %d", block_number);
+                                   break;
+                              }
+                         }
+                    }
+
+                    if ((last_block != -1) && (block_number > last_block)) {
                          state = S_END;
                          break;
                     }
+                    windowblock = 0;
+                    curr_sent_count++;
                     state = S_SEND_DATA;
                     break;
                case GET_ERROR:
@@ -952,10 +1153,12 @@ int tftpd_send_file(struct thread_data *data)
                          }
                     }
                     /* Got an ERROR from the current master client */
-                    Strncpy(string, tftphdr->th_msg, sizeof(string));
                     if (data->trace)
+                    {
+                         Strncpy(string, tftphdr->th_msg, sizeof(string));
                          logger(LOG_DEBUG, "received ERROR <code: %d, msg: %s>",
                                 ntohs(tftphdr->th_code), string);
+                    }
                     if (multicast)
                     {
                          logger(LOG_DEBUG, "Marking client as done");
@@ -999,12 +1202,14 @@ int tftpd_send_file(struct thread_data *data)
                          /* client is a new client structure */
                          sa =  &client_info->client;
                          /* nedd to send an oack to that client */
-                         state = S_SEND_OACK;                
+                         state = S_SEND_OACK;
                          fseek(fp, 0, SEEK_SET);
+			 /* reset the last block received counter */
+			 last_ackd_block = -1;
                     }
                     else
                     {
-                         logger(LOG_INFO, "No more client, end of tranfers");
+                         logger(LOG_INFO, "No more client, end of transfers");
                          fclose(fp);
                          return OK;
                     }
